"""Server class for API connections"""
import json
import importlib.util
import warnings
from typing import List, Dict, Optional, Any, IO, Tuple, Union, Iterator
from functools import wraps
import requests
from .utils import (request, build_portal_params, build_script_params,
                    filename_from_url, PlaceholderDict)
from .const import (PORTAL_PREFIX, FMSErrorCode, API_VERSIONS, API_PATH_PREFIX,
                    API_PATH)
from .exceptions import BadJSON, FileMakerError, RecordError
from .record import Record
from .foundset import Foundset


class Server(object):
    """The server class provides easy access to the FileMaker Data API

    Get an instance of this class, login, get a record, logout:

        import fmrest
        fms = fmrest.Server('https://server-address.com',
                    user='db user name',
                    password='db password',
                    database='db name',
                    layout='db layout',
                    api_version='v1'
                   )
        fms.login()
        fms.get_record(1)
        fms.logout()

    Or use as with statement, logging out automatically:

        with fms as my_server:
            my_server.login()
            # do stuff
    """

    def __init__(self, url: str, user: str,
                 password: str, database: str, layout: str,
                 data_sources: Optional[List[Dict]] = None,
                 verify_ssl: Union[bool, str] = True,
                 type_conversion: bool = False,
                 auto_relogin: bool = False,
                 proxies: Optional[Dict] = None,
                 api_version: Optional[str] = None) -> None:
        """Initialize the Server class.

        Parameters
        ----------
        url : str
            Address of the FileMaker Server, e.g. https://my-server.com or https://127.0.0.1
            Note: Data API must use https.
        user : str
            Username to log into your database
            Note: make sure it belongs to a privilege set that has fmrest extended privileges.
        password : str
            Password to log into your database
        database : str
            Name of database without extension, e.g. Contacts
        layout : str
            Layout to work with. Can be changed between calls by setting the layout attribute again,
            e.g.: fmrest_instance.layout = 'new_layout'.
        data_sources : list, optional
            List of dicts in formatj
                [{'database': 'db_file', 'username': 'admin', 'password': 'admin'}]
            Use this if for your actions you need to be authenticated to multiple DB files.
        verify_ssl : bool or str, optional
            Switch to set if certificate should be verified.
            Use False to disable verification. Default True.
            Use string path to a root cert pem file, if you work with a custom CA.
        type_conversion : bool, optional
            If True, attempt to convert string values into their potential original types.
            In previous versions of the FileMaker Data API only strings were returned and there was
            no way of knowing the correct type of a requested field value.

            Be cautious with this parameter, as results may be different from what you expect!

            Values will be converted into int, float, datetime, timedelta, string. This happens
            on a record level, not on a foundset level.
        auto_relogin : bool, optional
            If True, tries to automatically get a new token (re-login) when a
            request comes back with a 952 (invalid token) error. Defaults to
            False.
        proxies : dict, optional
            Pass requests through a proxy, configure like so:
            { 'https': 'http://127.0.0.1:8080' }
        api_version : str, optional
            Configure which version of the data API should be queried (e.g. v1)
            It is recommended to set the version explicitly to prevent
            potential future breaking changes.
        """

        self.url = url
        self.user = user
        self.password = password
        self.database = database
        self.layout = layout
        self.data_sources = [] if data_sources is None else data_sources
        self.verify_ssl = verify_ssl
        self.auto_relogin = auto_relogin
        self.proxies = proxies

        if not api_version:
            warnings.warn('No api_version given. Defaulting to v1.')
            self.api_version = 'v1'
        elif api_version not in API_VERSIONS:
            raise ValueError(f'Invalid API version. Choose one of {API_VERSIONS}')
        else:
            self.api_version = api_version

        self.type_conversion = type_conversion
        if type_conversion and not importlib.util.find_spec("dateutil"):
            warnings.warn('Turning on type_conversion needs the dateutil module, which '
                          'does not seem to be present on your system.')

        if url[:5] != 'https':
            raise ValueError('Please make sure to use https, otherwise calls to the Data '
                             'API will not work.')

        self._token: Optional[str] = None
        self._last_fm_error: Optional[int] = None
        self._last_script_result: Optional[Dict[str, List]] = None
        self._headers: Dict[str, str] = {}
        self._set_content_type()

    def __enter__(self) -> 'Server':
        return self

    def __exit__(self, exc_type, exc_val, exc_traceback) -> None:
        self.logout()

    def __repr__(self) -> str:
        return '<Server logged_in={} database={} layout={}>'.format(
            bool(self._token), self.database, self.layout
        )

    def _get_api_path(self, resource: str) -> str:
        resource_path = resource.split('.')
        if len(resource_path) > 1:
            if resource_path[0] == 'meta':
                path = API_PATH['meta'][resource_path[1]]
            else:
                raise ValueError('Invalid API path')
        else:
            path = API_PATH[resource_path[0]]

        return (API_PATH_PREFIX.format(version=self.api_version) +
                path.format_map(PlaceholderDict(database=self.database,
                                                layout=self.layout)))

    def _with_auto_relogin(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if not self.auto_relogin:
                return f(self, *args, **kwargs)

            try:
                return f(self, *args, **kwargs)
            except FileMakerError:
                if self.last_error == FMSErrorCode.INVALID_DAPI_TOKEN.value:
                    # got invalid token error; try to get a new token
                    self._token = None
                    self.login()
                    # ... now perform original request again
                    return f(self, *args, **kwargs)
                raise  # if another error occurred, re-raise the exception
        return wrapper

    def login(self) -> Optional[str]:
        """Logs into FMServer and returns access token.

        Authentication happens via HTTP Basic Auth. Subsequent calls to the API will then use
        the return session token.

        Note that OAuth is currently not supported.
        """

        path = self._get_api_path('auth').format(token='')
        data = {'fmDataSource': self.data_sources}

        response = self._call_filemaker('POST', path, data, auth=(self.user, self.password))
        self._token = response.get('token', None)

        return self._token

    def logout(self) -> bool:
        """Logs out of current session. Returns True if successful.

        Note: this method is also called by __exit__
        """

        # token is expected in endpoint for logout
        path = self._get_api_path('auth').format(token=self._token)

        # remove token, so that the Authorization header is not sent for logout
        # (_call_filemaker() will update the headers)
        self._token = ''
        self._call_filemaker('DELETE', path)

        return self.last_error == FMSErrorCode.SUCCESS.value

    def create(self, record: Record) -> Optional[int]:
        """Shortcut to create_record method. Takes record instance and calls create_record."""
        # TODO: support for handling foundset instances inside record instance
        return self.create_record(record.to_dict(ignore_portals=True, ignore_internal_ids=True))

    @_with_auto_relogin
    def create_record(self, field_data: Dict[str, Any],
                      portals: Optional[Dict[str, Any]] = None,
                      scripts: Optional[Dict[str, List]] = None) -> Optional[int]:
        """Creates a new record with given field data and returns new internal record id.

        Parameters
        -----------
        field_data : dict
            Dict of field names as defined in FileMaker: E.g.: {'name': 'David', 'drink': 'Coffee'}
        scripts : dict, optional
            Specify which scripts should run when with which parameters
            Example: {'prerequest': ['my_script', 'my_param']}
            Allowed types: 'prerequest', 'presort', 'after'
            List should have length of 2 (both script name and parameter are required.)
        portals : dict
            Specify the records that should be created via a portal (must allow creation of records)
            Example: {'my_portal': [
                {'TO::field': 'hello', 'TO::field2': 'world'},
                {'TO::field': 'another record'}
            ]
        """
        path = self._get_api_path('record')

        request_data: Dict = {'fieldData': field_data}
        if portals:
            request_data['portalData'] = portals

        # build script param object in FMSDAPI style
        script_params = build_script_params(scripts) if scripts else None
        if script_params:
            request_data.update(script_params)

        response = self._call_filemaker('POST', path, request_data)
        record_id = response.get('recordId')

        return int(record_id) if record_id else None

    def edit(self, record: Record, validate_mod_id: bool = False) -> bool:
        """Shortcut to edit_record method. Takes (modified) record instance and calls edit_record"""
        mod_id = record.modification_id if validate_mod_id else None
        return self.edit_record(record.record_id, record.modifications(), mod_id)

    @_with_auto_relogin
    def edit_record(self, record_id: int, field_data: Dict[str, Any],
                    mod_id: Optional[int] = None, portals: Optional[Dict[str, Any]] = None,
                    scripts: Optional[Dict[str, List]] = None) -> bool:
        """Edits the record with the given record_id and field_data. Return True on success.

        Parameters
        -----------
        record_id : int
            FileMaker's internal record id.
        field_data: dict
            Dict of field names as defined in FileMaker: E.g.: {'name': 'David', 'drink': 'Coffee'}

            To delete related records, use {'deleteRelated': 'Orders.2'}, where 2 is the record id
            of the related record.
        mod_id: int, optional
            Pass a modification id to only edit the record when mod_id matches the current mod_id of
            the server. This is only supported for records in the current table, not related
            records.
        portals : dict
            Specify the records that should be edited via a portal.
            If recordId is not specified, a new record will be created.
            Example: {'my_portal': [
                {'TO::field': 'hello', 'recordId': '42'}
            ]
        scripts : dict, optional
            Specify which scripts should run when with which parameters
            Example: {'prerequest': ['my_script', 'my_param']}
            Allowed types: 'prerequest', 'presort', 'after'
            List should have length of 2 (both script name and parameter are required.)
        """
        path = self._get_api_path('record_action').format(record_id=record_id)

        request_data: Dict = {'fieldData': field_data}
        if mod_id:
            request_data['modId'] = str(mod_id)

        if portals:
            request_data['portalData'] = portals

        # build script param object in FMSDAPI style
        script_params = build_script_params(scripts) if scripts else None
        if script_params:
            request_data.update(script_params)

        self._call_filemaker('PATCH', path, request_data)

        return self.last_error == FMSErrorCode.SUCCESS.value

    def delete(self, record: Record) -> bool:
        """Shortcut to delete_record method. Takes record instance and calls delete_record."""
        try:
            record_id = record.record_id
        except AttributeError:
            raise RecordError('Not a valid record instance. record_id is missing.') from None

        return self.delete_record(record_id)

    @_with_auto_relogin
    def delete_record(self, record_id: int, scripts: Optional[Dict[str, List]] = None):
        """Deletes a record for the given record_id. Returns True on success.

        Parameters
        -----------
        record_id : int
            FileMaker's internal record id.
        scripts : dict, optional
            Specify which scripts should run when with which parameters
            Example: {'prerequest': ['my_script', 'my_param']}
            Allowed types: 'prerequest', 'presort', 'after'
            List should have length of 2 (both script name and parameter are required.)
        """
        path = self._get_api_path('record_action').format(record_id=record_id)

        params = build_script_params(scripts) if scripts else None

        self._call_filemaker('DELETE', path, params=params)

        return self.last_error == FMSErrorCode.SUCCESS.value

    @_with_auto_relogin
    def get_record(self, record_id: int, portals: Optional[List[Dict]] = None,
                   scripts: Optional[Dict[str, List]] = None,
                   layout: Optional[str] = None) -> Record:
        """Fetches record with given ID and returns Record instance

        Parameters
        -----------
        record_id : int
            The FileMaker record id. Be aware that record ids CAN change (e.g. in cloned databases)
        portals : list
            A list of dicts in format [{'name':'objectName', 'offset':1, 'limit':50}]

            Use this if you want to limit the amout of data returned. Offset and limit are optional
            with default values of 1 and 50, respectively.
            All portals will be returned when portals==None. Default None.
        scripts : dict, optional
            Specify which scripts should run when with which parameters
            Example: {'prerequest': ['my_script', 'my_param']}
            Allowed types: 'prerequest', 'presort', 'after'
            List should have length of 2 (both script name and parameter are required.)
        layout : str, optional
            Passing a layout name allows you to set the response (!) layout.
            This is helpful, for example, if you want to limit the number of fields/portals being
            returned and have a dedicated response layout.
        """
        path = self._get_api_path('record_action').format(record_id=record_id)

        params = build_portal_params(portals, True) if portals else {}
        params['layout.response'] = layout

        # build script param object in FMSDAPI style
        script_params = build_script_params(scripts) if scripts else None
        if script_params:
            params.update(script_params)

        response = self._call_filemaker('GET', path, params=params)

        # pass response to foundset generator function. As we are only requesting one record though,
        # we only re-use the code and immediately consume the first (and only) record via next().
        return next(self._process_foundset_response(response))

    @_with_auto_relogin
    def perform_script(self, name: str,
                       param: Optional[str] = None) -> Tuple[Optional[int], Optional[str]]:
        """Performs a script with the given name and parameter.

        Returns tuple containing script error and result.

        Parameters:
        --------
        name : str
            The script name as defined in FileMaker Pro
        param: str
            Optional script parameter
        """
        path = self._get_api_path('script').format(script_name=name)

        response = self._call_filemaker('GET', path, params={'script.param': param})

        script_error = response.get('scriptError', None)
        script_error = int(script_error) if script_error else None
        script_result = response.get('scriptResult', None)

        return script_error, script_result

    @_with_auto_relogin
    def upload_container(self, record_id: int, field_name: str, file_: IO) -> bool:
        """Uploads the given binary data for the given record id and returns True on success.
        Parameters
        -----------
        record_id : int
            The FileMaker record id
        field_name : str
            Name of the container field on the current layout without TO name. E.g.: my_container
        file_ : fileobj
            File object as returned by open() in binary mode.
        """
        path = self._get_api_path('record_action').format(record_id=record_id)
        path += '/containers/' + field_name + '/1'

        # requests library handles content type for multipart/form-data incl. boundary
        self._set_content_type(False)
        self._call_filemaker('POST', path, files={'upload': file_})

        return self.last_error == FMSErrorCode.SUCCESS.value

    @_with_auto_relogin
    def get_records(self, offset: int = 1, limit: int = 100,
                    sort: Optional[List[Dict[str, str]]] = None,
                    portals: Optional[List[Dict[str, Any]]] = None,
                    scripts: Optional[Dict[str, List]] = None,
                    layout: Optional[str] = None) -> Foundset:
        """Requests all records with given offset and limit and returns result as
        (sorted) Foundset instance.

        Parameters
        -----------
        offset : int, optional
            Offset for the query, starting at 1, default 1
        limit : int, optional
            Limit the amount of returned records. Defaults to 100
        sort : list of dicts, optional
            A list of sort criteria. Example:
                [{'fieldName': 'name', 'sortOrder': 'descend'}]
        portals : list of dicts, optional
            Define which portals you want to include in the result.
            Example: [{'name':'objectName', 'offset':1, 'limit':50}]
            Defaults to None, which then returns all portals with default offset and limit.
        scripts : dict, optional
            Specify which scripts should run when with which parameters
            Example: {'prerequest': ['my_script', 'my_param']}
            Allowed types: 'prerequest', 'presort', 'after'
            List should have length of 2 (both script name and parameter are required.)
        layout : str, optional
            Passing a layout name allows you to set the response (!) layout.
            This is helpful, for example, if you want to limit the number of fields/portals being
            returned and have a dedicated response layout.
        """
        path = self._get_api_path('record')

        params = build_portal_params(portals, True) if portals else {}
        params['_offset'] = offset
        params['_limit'] = limit
        params['layout.response'] = layout

        if sort:
            params['_sort'] = json.dumps(sort)

        # build script param object in FMSDAPI style
        script_params = build_script_params(scripts) if scripts else None
        if script_params:
            params.update(script_params)

        response = self._call_filemaker('GET', path, params=params)
        info = response.get('dataInfo', {})

        return Foundset(self._process_foundset_response(response), info)

    @_with_auto_relogin
    def find(self, query: List[Dict[str, Any]],
             sort: Optional[List[Dict[str, str]]] = None,
             offset: int = 1, limit: int = 100,
             portals: Optional[List[Dict[str, Any]]] = None,
             scripts: Optional[Dict[str, List]] = None,
             layout: Optional[str] = None) -> Foundset:
        """Finds all records matching query and returns result as a Foundset instance.

        Parameters
        -----------
        query : list of dicts
            A list of find queries, specified as 'field_name': 'field_value'
            Example:
                [{'drink': 'Coffee'}, {'drink': 'Dr. Pepper'}] will find matches for either Coffee
                or Dr. Pepper.

                You can also negate find requests by adding a key "omit" with value "true".

                Generally, all FileMaker Pro operators are supported. So, wildcard finds with "*" or
                exact matches with "==" should all work like in Pro.
        sort : list of dicts, optional
            A list of sort criteria. Example:
                [{'fieldName': 'name', 'sortOrder': 'descend'}]
        offset : int, optional
            Offset for the query, starting at 1, default 1
        limit : int, optional
            Limit the amount of returned records. Defaults to 100
        portals : list of dicts, optional
            Define which portals you want to include in the result.
            Example: [{'name':'objectName', 'offset':1, 'limit':50}]
            Defaults to None, which then returns all portals with default offset and limit.
        scripts : dict, optional
            Specify which scripts should run when with which parameters
            Example: {'prerequest': ['my_script', 'my_param']}
            Allowed types: 'prerequest', 'presort', 'after'
            List should have length of 2 (both script name and parameter are required.)
        layout : str, optional
            Passing a layout name allows you to set the response (!) layout.
            Your find will still be performed based on the Server.layout attribute.
            This is helpful, for example, if you want to limit the number of fields/portals being
            returned and have a dedicated response layout.
        """
        path = self._get_api_path('find')

        data = {
            'query': query,
            'sort': sort,
            'limit': str(limit),
            'offset': str(offset),
            'layout.response': layout
        }

        # build script param object in FMSDAPI style
        script_params = build_script_params(scripts) if scripts else None
        if script_params:
            data.update(script_params)

        # build portal param object in FMSDAPI style
        portal_params = build_portal_params(portals) if portals else None
        if portal_params:
            data.update(portal_params)

        # FM Data API from v17 cannot handle null values, so we remove all Nones from data
        data = {k:v for k, v in data.items() if v is not None}

        response = self._call_filemaker('POST', path, data=data)
        info = response.get('dataInfo', {})

        return Foundset(self._process_foundset_response(response), info)

    def fetch_file(self, file_url: str,
                   stream: bool = False) -> Tuple[str, Optional[str], Optional[str], requests.Response]:
        """Fetches the file from the given url.

        Returns a tuple of filename (unique identifier), content type (e.g. image/png), length,
        and a requests response object. You can access contents by response.content.

        Example:
            url = record.container_field
            name, type_, length, content = fms.fetch_file(url)

        Parameters
        -----------
        file_url : str
            URL to file as returned by FMS.
            Example:
            https://address/Streaming_SSL/MainDB/unique-identifier.png?RCType=EmbeddedRCFileProcessor
        stream : bool, optional
            Set this to True if you don't want the file to immediately be loaded into memory.
            This let's you decide how you want to handle large files before downloading them.
            Access to headers is given before downloading.
            If you are not consuming all data, make sure to close the connection after use by
            calling response.close().
        """
        name = filename_from_url(file_url)
        response = request(method='get',
                           url=file_url,
                           verify=self.verify_ssl,
                           stream=stream,
                           proxies=self.proxies)

        return (name,
                response.headers.get('Content-Type'),
                response.headers.get('Content-Length'),
                response)

    @_with_auto_relogin
    def set_globals(self, globals_: Dict[str, Any]) -> bool:
        """Set global fields for the currently active session. Returns True on success.

        Global fields do not need to be placed on the layout and can be used for establishing
        relationships of which the global is a match field.

        Parameters
        -----------
        globals_ : dict
            Dict of { field name : value }
            Note that field names must be fully qualified, i.e. contain the TO name
            Example:
                { 'Table::myField': 'whatever' }
        """
        path = self._get_api_path('global')

        data = {'globalFields': globals_}

        self._call_filemaker('PATCH', path, data=data)
        return self.last_error == FMSErrorCode.SUCCESS.value

    @property
    def last_error(self) -> Optional[int]:
        """Returns last error number returned by FileMaker Server as int.

        Error is set by _call_filemaker method. If error == -1, the previous request failed
        and no FM error code is available. If no request was made yet, last_error will be None.
        """
        error: Optional[int]

        if self._last_fm_error:
            error = int(self._last_fm_error)
        else:
            error = None
        return error

    @property
    def last_script_result(self) -> Dict:
        """Returns last script results as returned by FMS as dict in format {type: [error, result]}

        Only returns keys that have a value from the last call. I.e. 'presort' will
        only be present if the last call performed a presort script.
        The returned error (0th element in list) will always be converted to int.
        """
        result: Dict = {}

        if self._last_script_result:
            result = {
                k:[int(v[0]), v[1]] for k, v in self._last_script_result.items() if v[0] is not None
            }
        return result

    def get_product_info(self) -> Dict:
        """Fetches product info and returns Dict instance

        Parameters
        -----------
        none
        """
        path = self._get_api_path('meta.product')

        response = self._call_filemaker('GET', path)

        return response.get('productInfo', None)

    def get_databases(self) -> Dict:
        """Fetches database names and returns Dict instance

        Parameters
        -----------
        none
        """
        path = self._get_api_path('meta.databases')

        # https://fmhelp.filemaker.com/docs/18/en/dataapi/#get-metadata_get-database-names
        # If Filter Databases in Client Applications is disabled:
        # no Authorization header is required.
        #
        # If Filter Databases in Client Applications is enabled:
        # Authorization: a base64-encoded string representing the account name
        # and password to use to log in to the hosted database.

        # See discussion here: https://github.com/davidhamann/python-fmrest/pull/46#issuecomment-1079328531
        # "This [will] eventually overwrite the Authorization header with the data provided in the auth 
        # argument (a comment there might be good as it could lead to confusion since we're adding an 
        # Authorization header, but requests adds one as well)."

        response = self._call_filemaker('GET', path, auth=(self.user, self.password))

        return response.get('databases', None)

    @_with_auto_relogin
    def get_layouts(self) -> Dict:
        """Fetches database layout names and returns Dict instance

        Parameters
        -----------
        none
        """
        path = self._get_api_path('meta.layouts')

        response = self._call_filemaker('GET', path)

        return response.get('layouts', None)

    @_with_auto_relogin
    def get_scripts(self) -> Dict:
        """Fetches database script names and returns Dict instance

        Parameters
        -----------
        none
        """
        path = self._get_api_path('meta.scripts')

        response = self._call_filemaker('GET', path)

        return response.get('scripts', None)

    @_with_auto_relogin
    def get_layout(self) -> Dict:
        """Fetches layout metadata and returns Dict instance

        Parameters
        -----------
        none
        """
        path = self._get_api_path('meta.layouts') + f'/{self.layout}'

        response = self._call_filemaker('GET', path)

        return response.get('fieldMetaData', None)

    def _call_filemaker(self, method: str, path: str,
                        data: Optional[Dict] = None,
                        params: Optional[Dict] = None,
                        **kwargs: Any) -> Dict:
        """Calls a FileMaker Server Data API path and returns the parsed fms response data

        Parameters
        -----------
        method : str
            The http request method, e.g. POST
        path : str
            The API path, /fmi/data/v1/databases/:database/...
        data : dict of str : str, optional
            Dict of parameter data for http request
            Can be None if API expects no data, e.g. for logout
        params : dict of str : str, optional
            Dict of get parameters for http request
            Can be None if API expects no params
        auth : tuple of str, str, optional
            Tuple containing user and password for HTTP basic
            auth
        """

        url = self.url + path
        request_data = json.dumps(data) if data else None

        # if we have a token, make sure it's included in the header
        # if not, the Authorization header gets removed (necessary for example
        # for logout)
        self._update_token_header()

        response = request(method=method,
                           headers=self._headers,
                           url=url,
                           data=request_data,
                           verify=self.verify_ssl,
                           params=params,
                           proxies=self.proxies,
                           **kwargs)

        try:
            response_data = response.json()
        except json.decoder.JSONDecodeError as ex:
            raise BadJSON(ex, response) from None

        fms_messages = response_data.get('messages')
        fms_response = response_data.get('response')

        self._update_script_result(fms_response)
        self._last_fm_error = fms_messages[0].get('code', -1)
        if self.last_error != FMSErrorCode.SUCCESS.value:
            raise FileMakerError(self._last_fm_error,
                                 fms_messages[0].get('message', 'Unkown error'))

        self._set_content_type() # reset content type

        return fms_response

    def _update_script_result(self, response: Dict) -> Dict[str, List]:
        """Extracts script result data from fms response and updates script result attribute"""
        self._last_script_result = {
            'prerequest': [
                response.get('scriptError.prerequest', None),
                response.get('scriptResult.prerequest', None)
            ],
            'presort': [
                response.get('scriptError.presort', None),
                response.get('scriptResult.presort', None)
            ],
            'after': [
                response.get('scriptError', None),
                response.get('scriptResult', None)
            ]
        }

        return self._last_script_result

    def _update_token_header(self) -> Dict[str, str]:
        """Update header to include access token (if available) for subsequent calls."""
        if self._token:
            self._headers['Authorization'] = 'Bearer ' + self._token
        else:
            self._headers.pop('Authorization', None)
        return self._headers

    def _set_content_type(self, type_: Union[str, bool] = 'application/json') -> Dict[str, str]:
        """Set the Content-Type header and returns the updated _headers dict.

        Parameters
        -----------
        type_ : str, boolean
            String definining the content type for the HTTP header or False to remove the
            Content-Type key from _headers (i.e. let the requests lib handle the Content-Type.)
        path : str
        """
        if isinstance(type_, str):
            self._headers['Content-Type'] = type_
        elif not type_:
            self._headers.pop('Content-Type')
        else:
            raise ValueError
        return self._headers

    def _process_foundset_response(self, response: Dict) -> Iterator[Record]:
        """Generator function that takes a response object, brings it into a Foundset/Record
        structure and yields processed Records.

        Lazily processing and yielding the results is slightly faster than building a list upfront
        when you deal with big foundsets containing records that each have many portal records.
        It won't save us much memory as we still hold the response, but initial processing time goes
        down, and we only need to build the records when we actually use them.
        (may think of another approach if it proves to be more pain than gain though)

        Parameters
        -----------
        response : dict
            FMS response from a _call_filemaker request
        """
        data = response['data']

        for record in data:
            field_data = record['fieldData']

            # Add meta fields to record.
            # TODO: this can clash with fields that have the same name. Find a better
            # way (maybe prefix?).
            # Note that portal foundsets have the recordId field included by default
            # (without the related table prefix).
            field_data['recordId'] = record.get('recordId')
            field_data['modId'] = record.get('modId')

            keys = list(field_data)
            values = list(field_data.values())

            portal_info = {}
            for entry in record.get('portalDataInfo', []):
                # a portal is identified by its object name, or, if not available, its TO name
                portal_identifier = entry.get('portalObjectName', entry['table'])
                portal_info[portal_identifier] = entry

            for portal_name, rows in record['portalData'].items():
                keys.append(PORTAL_PREFIX + portal_name)

                # further delay creation of portal record instances
                related_records = (
                    Record(list(row), list(row.values()),
                           in_portal=True, type_conversion=self.type_conversion
                          ) for row in rows
                )
                # add portal foundset to record
                values.append(Foundset(related_records, portal_info.get(portal_name, {})))

            yield Record(keys, values, type_conversion=self.type_conversion)



"""Record class for FileMaker record responses"""

from typing import List, Dict, Any, Optional
from .utils import convert_string_type
from .const import PORTAL_PREFIX

class Record(object):
    """A FileMaker record representation.

    (with ideas from: https://github.com/kennethreitz/records)
    """
    __slots__ = ('_keys', '_values', '_in_portal', '_modifications')

    def __init__(self, keys: List[str], values: List[Any],
                 in_portal: bool = False, type_conversion: bool = False) -> None:
        """Initialize the Record class.

        Parameters
        ----------
        keys : list
            List of keys (fields) for this Record as returned by FileMaker Server
        values : list
            Values corresponding to keys
        in_portal : bool
            If true, this record instance describes a related record from a portal. This is a
            special case as portal records are treated differently by the Data API and don't get
            all standard keys (modId is missing).
        type_conversion : bool, optional
            If True, attempt to convert string values into their potential original types.
            FileMaker Data API always returns strings and there is no way of knowing the correct
            type of a requested field value. Be cautious with this parameter!
            Values will be converted into int, float, datetime, timedelta, string.
        """

        self._keys = keys

        self._values: List[Any]
        if type_conversion:
            self._values = []
            for value in values:
                parsed = convert_string_type(value) if isinstance(value, str) else value
                self._values.append(parsed)
        else:
            self._values = values

        self._in_portal = in_portal
        self._modifications: Dict[str, Any] = {}

        if len(self._keys) != len(self._values):
            raise ValueError("Length of keys does not match length of values.")

    def __repr__(self) -> str:
        return '<Record id={} modification_id={} is_dirty={}>'.format(
            self.record_id,
            self.modification_id,
            self.is_dirty
        )

    def __getitem__(self, key: str) -> Any:
        """Returns value for given key. For dict lookups, like my_id = record['id']."""
        keys = self.keys()

        try:
            index = keys.index(key)
            return self.values()[index]
        except ValueError:
            raise KeyError(("No field named {}. Note that the Data API only returns fields "
                            "placed on your FileMaker layout.").format(key))

    def __getattr__(self, key: str) -> Any:
        """Returns value for given key. For attribute lookups, like my_id = record.id.

        Calls __getitem__ for key access.
        """
        try:
            return self[key]
        except KeyError as ex:
            raise AttributeError(ex) from None

    def __setitem__(self, key: str, value: Any) -> None:
        """Allows changing values of fields available in _keys.

        Modified keys land in _modifications and are later used to write values back to
        FileMaker.
        """
        if key not in self.__slots__:
            # objects in __slots__ are the only allowed attributes.
            # all others are handled here
            if key not in self.keys():
                raise KeyError(str(key) + " is not a valid field name.")
            elif key.startswith(PORTAL_PREFIX):
                raise KeyError(
                    ("Portal data cannot be set through the record instance. "
                     "To edit portal data, build a dict and pass it to edit_records().")
                )
            elif value != self[key]:
                # store modified key and value for later re-use
                self._modifications[key] = value

                # also update the value in _values, so that values() returns expected data
                index = self.keys().index(key)
                self._values[index] = value
        else:
            # allow setting of attributes in __slots__
            super().__setattr__(key, value)

    def __setattr__(self, key: str, value: Any) -> None:
        """See __setitem__. Returns AttributeError if trying to set a value for a field/attribute
        not existing in the record instance.
        """
        try:
            return self.__setitem__(key, value)
        except KeyError as ex:
            raise AttributeError(ex) from None

    def modifications(self) -> Dict[str, Any]:
        """Returns a dict of changed keys in the form of {key : new_value}.

        Used for writing back record changes via Server.edit(record).
        """
        return self._modifications

    @property
    def is_dirty(self) -> bool:
        """Returns True if key values have been modified."""
        return len(self._modifications) > 0

    @property
    def record_id(self) -> int:
        """Returns the internal record id.

        This is exposed as a method to reliably return the record id, even if the API might change
        the field name in the future.
        """
        return int(self.recordId)

    @property
    def modification_id(self) -> Optional[int]:
        """Returns the internal modification id.

        This is exposed as a method to reliably return the modification id, even if the API might
        change the field name in the future.
        """
        return None if self._in_portal else int(self.modId)

    def keys(self) -> List[str]:
        """Returns all keys of this record."""
        return self._keys

    def values(self) -> List[Any]:
        """Returns all values of this record."""
        return self._values

    def to_dict(self, ignore_portals: bool = False, ignore_internal_ids: bool = False) -> Dict[str, Any]:
        """Returns record values as dictionary of key: val."""
        zipped = zip(self.keys(), self.values())

        if ignore_portals:
            out = {k:v for k, v in zipped if not k.startswith(PORTAL_PREFIX)}
        else:
            out = dict(zipped)

        if ignore_internal_ids:
            out.pop('recordId', None)
            out.pop('modId', None)
        return out

    def pop(self, key: str, default: Any = None) -> Any:
        """Pops the record's key. Returns key's value or default."""
        keys = self.keys()

        try:
            value = self[key]
            index = keys.index(key)
            self._keys.pop(index)
            self._values.pop(index)
            return value
        except (KeyError, ValueError):
            return default



from typing import Optional, Any
import requests

class FMRestException(Exception):
    """The base fmrest Exception."""

class RequestException(FMRestException):
    """Exception for http request errors

    Re-raised after requests module exception
    """

    def __init__(self, original_exception: Exception, request_args: Any, request_kwargs: Any) -> None:
        """Parameters
        ----------
        original_exception
            The original exception raised by requests module
        request_args
            Args to the request function
        request_kwargs
            Keyword args to the request function
        """

        self._original_exception = original_exception
        self._request_args = request_args
        self._request_kwargs = request_kwargs
        super().__init__('Request error: {}'.format(original_exception))

class ResponseException(FMRestException):
    """Exception for http response errors

    Re-raised after requests module exception
    """

    def __init__(self, original_exception: Exception, response: requests.Response) -> None:
        """Parameters
        ----------
        original_exception
            The original exception raised by requests module
        response:
            Response object of requests module
        """
        self._response = response
        super().__init__(
            '{}, {} http response, content-type: {}'.format(
                original_exception,
                self._response.status_code,
                self._response.headers.get('content-type', None))
        )

class BadJSON(ResponseException):
    """Invalid json response"""

class FileMakerError(FMRestException):
    """Error raised by FileMaker Data API"""

    def __init__(self, error_code: Optional[int], error_message: str) -> None:
        super().__init__('FileMaker Server returned error {}, {}'.format(error_code, error_message))

class RecordError(FMRestException):
    """Error with the local Record instance."""



"""Utility functions for fmrest"""
from typing import List, Dict, Any, Iterator
import requests
from .exceptions import RequestException
from .const import TIMEOUT


def request(*args, **kwargs) -> requests.Response:
    """Wrapper around requests library request call"""
    try:
        return requests.request(*args, timeout=TIMEOUT, **kwargs)
    except Exception as ex:
        raise RequestException(ex, args, kwargs) from None

def build_portal_params(portals: List[Dict], names_as_string: bool = False) -> Dict[str, Any]:
    """Takes a list of dicts and returns a dict in a format as FMServer expects it.

    FMS expects portals and their options to be specified in the following format:
        portal=["Portal1", "Portal2"]&_offset.Portal1=1&_limit.Portal1=10

    This function will return a dict of params suitable for the requests module.

    Parameters
    -----------
    portals : list
        List of dicts with keys name, offset, limit
        Example:
            [
                {
                    'name': 'addressPortal', # FM object name of portal
                    'offset': 1,
                    'limit': 50
                },
                {
                    'name': 'notesPortal',
                    'offset': 1,
                    'limit': 50
                }
            ]
    names_as_string : bool
        For GET params the list of portal names might need to be turned into
        a string. Use this to get something like '["Portal1", "Portal2"]'.
        If False, portals key will have a value of type list, like ["Portal1", "Portal2"]
    """

    params: Dict[str, Any]

    portal_selector = [portal['name'] for portal in portals]
    if names_as_string:
        portal_param = "[" + ', '.join(map(lambda x: '"' + x + '"', portal_selector)) + "]"
        params = {'portal': portal_param}
        param_prefix = '_' # for GET we need an underscore as prefix
    else:
        params = {'portal': portal_selector}
        param_prefix = ''

    for portal in portals:
        params[param_prefix + 'offset.' + portal['name']] = portal.get('offset', 1)
        params[param_prefix + 'limit.' + portal['name']] = portal.get('limit', 50)

    return params

def build_script_params(scripts: Dict[str, List]) -> Dict[str, str]:
    """Takes simplified scripts object and returns scripts objects as FMSDAPI expects it.

    Example simplified scripts object:

        scripts = {
            'prerequest': ['script_name', 'script_parameter'],
            'presort': ['script_name', 'script_parameter'],
            'after': ['script_name', 'script_parameter']
        }
    """

    # read user-defined script configuration
    # if no values were passed, initilize list with None as script and param
    presort = scripts.get('presort', [None]*2)
    prerequest = scripts.get('prerequest', [None]*2)
    after = scripts.get('after', [None]*2)

    # format for FMSDAPI
    verbose_scripts = {
        'script.prerequest': prerequest[0],
        'script.prerequest.param': prerequest[1],
        'script.presort': presort[0],
        'script.presort.param': presort[1],
        'script': after[0],
        'script.param': after[1],
    }

    # return only keys that have a real value
    return {k:v for k, v in verbose_scripts.items() if v is not None}

def cache_generator(iterator: Iterator, cache: List) -> Iterator:
    """Takes iterator and cache list, caches values before yielding them.
    Eventually flagging cache as complete.

    Parameters
    ----------
    iterator : generator
        Generator to consume
    cache : list
        List holding list of cached values and state of caching.
        If cache[1] is True, all values have been cached.
    """

    for val in iterator:
        cache[0].append(val)
        yield val

    cache[1] = True # all values have been cached

def filename_from_url(url: str) -> str:
    """Returns filename from given remote container url."""

    # remove query string
    url = url.split('?')[0]

    # get last part of url. FileMaker Data API does not always include the file extension
    # (e.g. for non audio/image/video) so we cannot rely on matching for a filename with extension
    filename = url.split('/')[-1]

    return filename

def convert_string_type(value):
    """Quick and dirty way to convert strings into their (guessed) original type.

    FileMaker Data API only returns strings. Hopefully, we can throw this function away as
    soon as the Data API is out of beta :-)

    Not used when running fmrest with default parameters as returned values can
    be unexpected.
    """

    # int and float
    for type_ in int, float:
        try:
            return type_(value)
        except ValueError:
            pass

    # datetime / timedelta
    try:
        from dateutil.parser import parse
        from datetime import timedelta
        parsed = parse(value)
        if '/' not in value:
            #assume time, as FM always returns / for date and ts, and parse didn't raise ValueError
            parsed = timedelta(hours=parsed.hour, minutes=parsed.minute, seconds=parsed.second)
        return parsed
    except ValueError:
        pass

    # timedelta (above try will fail for >24h)
    time_split = value.split(':')
    if len(time_split) == 3:
        try:
            hours = int(time_split[0])
            minutes = int(time_split[1])
            seconds = int(time_split[2])
            return timedelta(hours=hours, minutes=minutes, seconds=seconds)
        except ValueError:
            pass

    # fall back to string
    return value


class PlaceholderDict(dict):
    """Used to allow missing values in format maps and keep placeholders intact
    Example:
        > '{a}, {b}'.format_map(PlaceholderDict(a='hello'))
        'hello, {b}'
    Also see: https://docs.python.org/3/library/stdtypes.html#str.format_map
    """
    def __missing__(self, key):
        return key.join('{}')



"""Foundset class for collections of Records"""

import itertools
from typing import Iterator, List, Any, Dict
from .utils import cache_generator
from .record import Record

class Foundset(object):
    """A set of Record instances

    Foundsets are used for both find results and portal data (related records)
    """
    def __init__(self, records: Iterator, info: Dict = {}) -> None:
        """Initialize the Foundset class.

        The foundset is cached while being consumed, so that subsequent iterations are possible.

        Parameters
        ----------
        records : generator
            Generator of Record instances representing the records in a foundset or
            related records from a portal.
        info : dictionary
            Dictionary of information about the foundset. This is 1:1 the dictionary that
            is delivered by FMS for any foundset.
        """
        self._records = records
        self._consumed = False
        self._info = info

        # We hold the list of cached values and the state of completion in a list
        # idea: https://codereview.stackexchange.com/a/178780/151724
        self._cache: List[Any] = [[], False]

        # cache_generator will yield the values and handle the caching
        self._iter = cache_generator(self._records, self._cache)

    def __iter__(self) -> Iterator:
        """Make foundset iterable.

        Returns iter for list of records already consumed from generator, or a chained object
        of cache list plus still-to-consume records. This makes sure foundsets can be properly used
        as a list.
        """

        if self._cache[1]:
            # all values have been cached
            return iter(self._cache[0])

        return itertools.chain(self._cache[0], self._iter)

    def __getitem__(self, index: int) -> Record:
        """Return item at index in the iterator. If it's already cached, then return cached version.
        Otherwise consume until found.

        Parameters
        ----------
        index : int
        """
        while index >= len(self._cache[0]):
            try:
                next(self._iter)
            except StopIteration:
                break

        return self._cache[0][index]

    def __repr__(self) -> str:
        return '<Foundset consumed_records={} is_complete={}>'.format(
            len(self._cache[0]), self.is_complete
        )

    @property
    def is_complete(self) -> bool:
        """Returns True if all values have been consumed. Otherwise False."""
        return self._cache[1]

    @property
    def info(self) -> Dict:
        """Returns data that is contained in the dataInfo section of the FMS response."""
        return self._info

    def to_df(self):
        """Returns a Pandas DataFrame of the Foundset. Must have Pandas installed.

        Note that portal data is not returned as part of the DataFrame.
        """
        try:
            import pandas as pd
        except ImportError as ex:
            raise Exception(
                "You need to have Pandas installed to use this feature. "
                "You can install it like this: 'pip install pandas'"
            ) from ex

        return pd.DataFrame(
            [r.to_dict(ignore_portals=True) for r in self]
        )



"""fmrest constants"""
import os
from enum import Enum, unique
from pkg_resources import get_distribution
from typing import Dict, Any

__version__ = get_distribution('python-fmrest').version

PORTAL_PREFIX = 'portal_'
TIMEOUT = int(os.environ.get('fmrest_timeout', 10))

API_VERSIONS = ('v1', 'v2', 'vLatest')
API_PATH_PREFIX = '/fmi/data/{version}'
API_PATH: Dict[str, Any] = {
    'meta': {
        'product':      '/productInfo',
        'databases':    '/databases',
        'layouts':      '/databases/{database}/layouts',
        'scripts':      '/databases/{database}/scripts'
    },
    'auth':             '/databases/{database}/sessions/{token}',
    'record':           '/databases/{database}/layouts/{layout}/records',
    'record_action':    '/databases/{database}/layouts/{layout}/records/{record_id}',
    'find':             '/databases/{database}/layouts/{layout}/_find',
    'script':           '/databases/{database}/layouts/{layout}/script/{script_name}',
    'global':           '/databases/{database}/globals'
}


@unique
class FMSErrorCode(Enum):
    """FMS error codes that are being referenced in the code"""
    SUCCESS = 0
    RECORD_MISSING = 101
    INVALID_USER_PASSWORD = 212
    INVALID_DAPI_TOKEN = 952



from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as ld:
    long_description = ld.read()

setup(
    name='python-fmrest',
    version='1.5.2',
    python_requires='>=3.6',
    author='David Hamann',
    author_email='dh@davidhamann.de',
    description='python-fmrest is a wrapper around the FileMaker Data API.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/davidhamann/python-fmrest',
    packages=['fmrest'],
    include_package_data=True,
    install_requires=['requests>=2'],
    extras_require={
        'cloud': ['pycognito>=0.1.4']
    },
    classifiers=(
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    )
)



"""Subclass of Server, specifically for connecting via the 'new' FileMaker Cloud"""
from typing import Dict, List, Optional, Union

try:
    import pycognito
except ImportError:
    _has_pycognito = False
else:
    _has_pycognito = True

from .const import API_PATH
from .server import Server

class CloudServer(Server):
    """The CloudServer class provides access to a FileMaker Cloud database. This requires authenticating with a
     Claris ID via Amazon Cognito, as described in the FileMaker 19 Data API Guide:

     https://help.claris.com/en/data-api-guide/

     Usage is otherwise identical to that of the standard fmrest Server class.

     Please note that the values for cognito_userpool_id and cognito_client_id are the same for all FileMaker Cloud
     connections at the time of writing, and thus have been provided as default values. These may, however, change in
     the future, in which case new values may be obtained per instructions here:

     https://help.claris.com/en/customer-console-help/content/create-fmid-token.html
     """

    def __init__(self,
                 url: str,
                 user: str,
                 password: str,
                 database: str,
                 layout: str,
                 cognito_userpool_id: str = 'us-west-2_NqkuZcXQY',
                 cognito_client_id: str = '4l9rvl4mv5es1eep1qe97cautn',
                 data_sources: Optional[List[Dict]] = None,
                 verify_ssl: Union[bool, str] = True,
                 type_conversion: bool = False,
                 auto_relogin: bool = False,
                 proxies: Optional[dict] = None,
                 cognito_proxies: Optional[dict] = None,
                 api_version: Optional[str] = None) -> None:
        """Initialize the CloudServer class.

        Parameters
        ----------
        url : str
            Address of the FileMaker Cloud Server, e.g. https://myteamname.account.filemaker-cloud.com
            Note: Data API must use https.
        user : str
            Claris ID name to log into your database
            Note: make sure it belongs to a privilege set that has fmrest extended privileges.
        password : str
            Claris ID Password to log into your database
        database : str
            Name of database without extension, e.g. Contacts
        layout : str
            Layout to work with. Can be changed between calls by setting the layout attribute again,
            e.g.: fmrest_instance.layout = 'new_layout'.
        cognito_userpool_id: str
            Amazon Cognito Userpool ID (required for FileMaker Cloud authentication).
            See https://help.claris.com/en/customer-console-help/content/create-fmid-token.html
            At time of writing, this value should be 'us-west-2_NqkuZcXQY'
        cognito_client_id:
            Amazon Cognito Client ID (required for FileMaker Cloud authentication).
            See https://help.claris.com/en/customer-console-help/content/create-fmid-token.html
            At time of writing, this value should be '4l9rvl4mv5es1eep1qe97cautn'
        data_sources : list, optional
            List of dicts in format
                [{'database': 'db_file', 'username': 'admin', 'password': 'admin'}]
            Use this if for your actions you need to be authenticated to multiple DB files.
        verify_ssl : bool or str, optional
            Switch to set if certificate should be verified.
            Use False to disable verification. Default True.
            Use string path to a root cert pem file, if you work with a custom CA.
        type_conversion : bool, optional
            If True, attempt to convert string values into their potential original types.
            In previous versions of the FileMaker Data API only strings were returned and there was
            no way of knowing the correct type of a requested field value.

            Be cautious with this parameter, as results may be different from what you expect!

            Values will be converted into int, float, datetime, timedelta, string. This happens
            on a record level, not on a foundset level.
        auto_relogin : bool, optional
            If True, tries to automatically get a new token (re-login) when a
            request comes back with a 952 (invalid token) error. Defaults to
            False.
        proxies : dict, optional
            Pass requests to the FileMaker Data API through a proxy. Proxies should be configured as follows:
            {'http': 'http://username:password@hostname:port', 'https': 'http://username:password@hostname:port'}
        cognito_proxies : dict, optional
            Proxy to use when authenticating the Claris ID with Cognito. When deployed to a server, it may be necessary
            to use a static proxy during Cognito authentication to avoid entering a sign-in verification loop
        """

        if not _has_pycognito:
            raise ImportError(
                'Please install pycognito for Claris Cloud support. '
                'You can do so with: pip install python-fmrest[cloud]'
            )

        super().__init__(url,
                         user,
                         password,
                         database,
                         layout,
                         data_sources=data_sources,
                         verify_ssl=verify_ssl,
                         type_conversion=type_conversion,
                         auto_relogin=auto_relogin,
                         api_version=api_version,
                         proxies=proxies)

        self.cognito_userpool_id = cognito_userpool_id
        self.cognito_client_id = cognito_client_id
        self.cognito_proxies = cognito_proxies
        self._fmid_token = None

    def _get_cognito_token(self) -> str:
        """Use Pycognito library to authenticate with Amazon Cognito and retrieve FMID token."""
        kwargs = {
            'user_pool_id': self.cognito_userpool_id,
            'client_id': self.cognito_client_id,
            'username': self.user
        }
        if self.cognito_proxies:
            kwargs['botocore_config'] = pycognito.boto3.session.Config(
                proxies=self.cognito_proxies
            )

        user = pycognito.Cognito(**kwargs)
        user.authenticate(self.password)
        return user.id_token

    def _get_bearer_token(self) -> Optional[str]:
        """Retrieve the bearer token needed to authenticate FileMaker Data API calls."""

        path = self._get_api_path('auth').format(token='')
        data = {'fmDataSource': self.data_sources}

        response = self._call_filemaker('POST', path, data=data)
        return response.get('token', None)

    def _update_token_header(self) -> Dict[str, str]:
        """Override token header update method to use FMID token if set."""

        if self._fmid_token:
            self._headers['Authorization'] = 'FMID ' + self._fmid_token
        elif self._token:
            self._headers['Authorization'] = 'Bearer ' + self._token
        else:
            self._headers.pop('Authorization', None)
        return self._headers

    def login(self) -> Optional[str]:
        """Override Server login so we obtain FMID token from Cognito first"""

        self._fmid_token = self._get_cognito_token()
        self._token = self._get_bearer_token()
        self._fmid_token = None  # Reset FMID token so auth headers are set appropriately for data api calls
        return self._token
